__all__ = ['ValidationError']

from pydantic import ValidationError as PydanticError


class ValidationError(PydanticError):
    def __init__(self, e: PydanticError):
        super(ValidationError, self).__init__(errors=e.raw_errors, model=e.model, )

    pass



# converters
#
# check Werkzeug builtin_converters in
# https://werkzeug.palletsprojects.com/en/0.15.x/routing/#builtin-converters

import typing as t


def convert_any(*args, **kwargs):
    """
    Handle converter type "any"
    :param args:
    :param kwargs:
    :return: return schema dict
    """
    schema = {
        'type': 'array',
        'items': {
            'type': 'string',
            'enum': args,
        }
    }
    return schema


def int_schema(*args, **kwargs):
    """
    Handle converter type "int"
    :param args:
    :param kwargs:
    :return: return schema dict
    """
    schema = {
        'type': 'integer',
        'format': 'int32',
    }
    if 'max' in kwargs:
        schema['maximum'] = kwargs['max']
    if 'min' in kwargs:
        schema['minimum'] = kwargs['min']
    return schema


def float_schema(*args, **kwargs):
    """
    Handle converter type "float"
    :param args:
    :param kwargs:
    :return: return schema dict
    """
    schema = {
        'type': 'number',
        'format': 'float',
    }
    return schema


def uuid_schema(*args, **kwargs):
    """
    Handle converter type "uuid"
    :param args:
    :param kwargs:
    :return: return schema dict
    """
    schema = {
        'type': 'string',
        'format': 'uuid',
    }
    return schema


def convert_path(*args, **kwargs):
    """
    Handle converter type "path"
    :param args:
    :param kwargs:
    :return: return schema dict
    """
    schema = {
        'type': 'string',
        'format': 'path',
    }
    return schema


def string_schema(*args, **kwargs):
    """
    Handle converter type "string"
    :param args:
    :param kwargs:
    :return: return schema dict
    """
    schema = {
        'type': 'string',
    }
    for prop in ['length', 'maxLength', 'minLength', 'maxlength']:
        if prop in kwargs:
            schema["minLength"] = kwargs[prop]
    return schema


def convert_default(*args, **kwargs):
    """
    Handle converter type "default"
    :param args:
    :param kwargs:
    :return: return schema dict
    """
    schema = {'type': 'string'}
    return schema


class BaseSchema:
    def __init__(self, _type: str, _format: str = None, *args: t.Any, **kwargs: t.Any) -> None:
        self.type = _type
        self.format = _format or _type

    def schema_json(self):
        return self.__dict__


class StringSchema(BaseSchema):
    def __init__(self, _format: str = None, *args: t.Any, **kwargs: t.Any):
        if 'length' in kwargs:
            self.maxLength = kwargs['length']
            self.minLength = kwargs['length']
        if 'maxlength' in kwargs:
            self.maxLength = kwargs['maxlength']
        if 'minlength' in kwargs:
            self.minLength = kwargs['minlength']
        super().__init__("string", _format)


class NumberSchema(BaseSchema):
    def __init__(self, _type, _format, *args: t.Any, **kwargs: t.Any):
        if 'max' in kwargs:
            self.maximum = kwargs['max']
        if 'min' in kwargs:
            self.minimum = kwargs['min']
        assert _type in ("integer", "number")
        super().__init__(_type, _format=_format)


class IntegerSchema(NumberSchema):
    def __init__(self, *args: t.Any, **kwargs: t.Any):
        super().__init__("integer", _format="int64", *args, **kwargs)


class FloatSchema(NumberSchema):
    def __init__(self, *args: t.Any, **kwargs: t.Any):
        super().__init__("number", _format="number", *args, **kwargs)


class UUIDSchema(StringSchema):
    def __init__(self, *args: t.Any, **kwargs: t.Any):
        super().__init__(_format="uuid", *args, **kwargs)


class PathSchema(StringSchema):
    def __init__(self, *args: t.Any, **kwargs: t.Any):
        super().__init__(_format="path", *args, **kwargs)


class AnySchema(StringSchema):
    def __init__(self, *args: t.Any, **kwargs: t.Any):
        self.enum = args
        super().__init__(_format="enum", **kwargs)


SCHEMAS: t.Mapping[str, t.Type[BaseSchema]] = {
    'any': AnySchema,
    'int': IntegerSchema,
    'float': FloatSchema,
    'uuid': UUIDSchema,
    'path': PathSchema,
    'string': StringSchema,
    'default': StringSchema
}


def get_converter(converter: str, *args, **kwargs):
    """
    Get conveter method from converter map
    :param converter: str: converter type
    :param args:
    :param kwargs:
    :return: return schema dict
    """
    return SCHEMAS[converter](*args, **kwargs)



import inspect
import re
from typing import Type, List, Mapping, Any, Tuple

from pydantic import BaseModel
from pydantic.typing import Literal
from werkzeug.datastructures import MultiDict

from . import schema


def convert_query_params(query_prams: MultiDict, model: Type[BaseModel]) -> dict:
    """
    :param query_prams: flask request.args
    :param model: query parameter's model
    :return resulting parameters
    """
    return {
        **query_prams.to_dict(),
        **{key: value for key, value in query_prams.to_dict(flat=False).items() if
           key in model.__fields__ and model.__fields__[key].is_complex()}
    }


def parse_path_params(route: str) -> Tuple[str, List]:
    """
    解析路径参数
    :param route: /hello/<int(min=2):age>
    :return

    path: /hello/{age}
    parameters: [{
                        "in": "path",
                        "name": "age",
                        "required": true,
                        "schema": {
                            "format": "int64",
                            "type": "integer"
                        }
                }],
    """
    subs = []
    parameters = []

    # /hello/<int(min=2):age>
    # 会转成两个元素
    # 0:None, None, hello
    # 1:int, min=2, age
    for converter, arguments, variable in parse_rule(route):
        if converter is None:
            subs.append(variable)
            continue
        subs.append(f'{{{variable}}}')

        args, kwargs = [], {}

        if arguments:
            args, kwargs = parse_converter_args(arguments)

        path_param_schema = get_path_param_schema(converter, *args, **kwargs).schema_json()

        parameters.append({
            'name': variable,
            'in': 'path',
            'required': True,
            'schema': path_param_schema,
        })
    path = ''.join(subs)
    return path, parameters


def parse_other_params(
        location: Literal["query", "header", "cookie"],
        model: Mapping[str, Any],
) -> List[Mapping[str, Any]]:
    params = []
    for name, _schema in model["properties"].items():
        params.append(
            {
                "name": name,
                "in": location,
                "schema": _schema,
                "required": name in model.get("required", []),
                "description": _schema.get("description", ""),
            }
        )
    return params


def get_operation_summary(func) -> str:
    """
    return a summary for operation in the method
    :param func:flask view function
    :return:str
    """
    # print(func)
    return func.summary or func.__qualname__.replace(".", " ").replace("_", " ").title()


def get_operation_description(func) -> str:
    """
    :param func: flask view function
    :return:str
    """
    return func.description or (inspect.getdoc(func) or "").split("\f")[0]


def get_path_param_schema(converter: str, *args, **kwargs) -> schema.BaseSchema:
    """
    获取路径参数对应的schema
    将werkzeug的转换器转换成openapi所需的schema

    例如路由：Rule('/<string(length=10):lang_code>')
    转换成schema
    "schema": {
              "format": "string",
              "maxLength": 10,
              "minLength": 10,
              "type": "string"
            }
    :param converter: str: converter type  'any'|'int'|'float'|'uuid'|'path'|'string'|'default'
    :param args:
    :param kwargs:
    """
    return schema.SCHEMAS[converter](*args, **kwargs)


_rule_re = re.compile(
    r"""
    (?P<static>[^<]*)                           # static rule data
    <
    (?:
        (?P<converter>[a-zA-Z_][a-zA-Z0-9_]*)   # converter name
        (?:\((?P<args>.*?)\))?                  # converter arguments
        \:                                      # variable delimiter
    )?
    (?P<variable>[a-zA-Z_][a-zA-Z0-9_]*)        # variable name
    >
    """,
    re.VERBOSE,
)


def parse_rule(rule):
    """Parse a rule and return it as generator. Each iteration yields tuples
    in the form ``(converter, arguments, variable)``. If the converter is
    `None` it's a static url part, otherwise it's a dynamic one.

    :internal:
    """
    pos = 0
    end = len(rule)
    do_match = _rule_re.match
    used_names = set()
    while pos < end:
        m = do_match(rule, pos)
        if m is None:
            break
        data = m.groupdict()
        if data["static"]:
            yield None, None, data["static"]
        variable = data["variable"]
        converter = data["converter"] or "default"
        if variable in used_names:
            raise ValueError("variable name %r used twice." % variable)
        used_names.add(variable)
        yield converter, data["args"] or None, variable
        pos = m.end()
    if pos < end:
        remaining = rule[pos:]
        if ">" in remaining or "<" in remaining:
            raise ValueError("malformed url rule: %r" % rule)
        yield None, None, remaining


def parse_converter_args(argstr):
    argstr += ","
    args = []
    kwargs = {}

    for item in _converter_args_re.finditer(argstr):
        value = item.group("stringval")
        if value is None:
            value = item.group("value")
        value = _pythonize(value)
        if not item.group("name"):
            args.append(value)
        else:
            name = item.group("name")
            kwargs[name] = value

    return tuple(args), kwargs


_converter_args_re = re.compile(
    r"""
    ((?P<name>\w+)\s*=\s*)?
    (?P<value>
        True|False|
        \d+.\d+|
        \d+.|
        \d+|
        [\w\d_.]+|
        [urUR]?(?P<stringval>"[^"]*?"|'[^']*')
    )\s*,
    """,
    re.VERBOSE | re.UNICODE,
)

_PYTHON_CONSTANTS = {"None": None, "True": True, "False": False}


def _pythonize(value):
    if value in _PYTHON_CONSTANTS:
        return _PYTHON_CONSTANTS[value]
    for convert in int, float:
        try:
            return convert(value)
        except ValueError:
            pass
    if value[:1] == value[-1:] and value[0] in "\"'":
        value = value[1:-1]
    return str(value)



import copy
from collections import defaultdict
from typing import Dict, List, Any

from flask import Flask

from . import utils


def generate_openapi(title: str,
                     version: str,
                     openapi_version: str,
                     app: Flask,
                     models: Dict[str, Dict],
                     description: str = None) -> Dict[str, Any]:
    """
    :param title:
    :param version:
    :param openapi_version:
    :param app:
    :param models:
    :param description:
    """

    routes: Dict[str:Dict] = dict()
    tags: Dict[str:Dict] = dict()
    groups: Dict[str:List] = defaultdict(list)
    for rule in app.url_map.iter_rules():
        # 视图函数
        func = old_func = app.view_functions[rule.endpoint]
        path, parameters = utils.parse_path_params(str(rule))
        methods = rule.methods
        for method in methods:
            if method in ['HEAD', 'OPTIONS']:
                continue
            if getattr(old_func, "view_class", None):
                cls = getattr(old_func, "view_class")
                func = getattr(cls, method.lower(), None)
            # 只有被siwadoc装饰了函数才加入openapi
            if not getattr(func, '_decorated', None):
                continue
            if not hasattr(func, 'tags'):
                func.tags = ['default']
            if not hasattr(func, 'group'):
                func.group = ''

            func_group = getattr(func, 'group', "")
            func_tags = [tag if tag != 'default' else func_group + "/" + tag for tag in
                         getattr(func, 'tags', ['default'])]

            groups[func_group].extend(func_tags)
            tags.update({tag: {"name": tag} for tag in func_tags})
            operation = {
                'summary': utils.get_operation_summary(func),
                'description': utils.get_operation_description(func),
                'operationID': func.__name__ + '__' + method.lower(),
                'tags': func_tags,
            }

            if hasattr(func, 'body'):
                operation['requestBody'] = {
                    'content': {
                        'application/json': {
                            'schema': {
                                '$ref': f'#/components/schemas/{func.body}'
                            }
                        }
                    }
                }
            parameters = copy.deepcopy(parameters)
            if hasattr(func, 'query'):
                parameters.extend(utils.parse_other_params('query', models[func.query]))
            if hasattr(func, 'header'):
                parameters.extend(utils.parse_other_params('header', models[func.header]))
            if hasattr(func, 'cookie'):
                parameters.extend(utils.parse_other_params('cookie', models[func.cookie]))
            operation['parameters'] = parameters

            operation['responses'] = {}
            has_2xx = False
            if hasattr(func, 'x'):
                for code, msg in func.x.items():
                    if code.startswith('2'):
                        has_2xx = True
                    operation['responses'][code] = {
                        'description': msg,
                    }

            if hasattr(func, 'resp'):
                operation['responses']['200'] = {
                    'description': 'Successful Response',
                    'content': {
                        'application/json': {
                            'schema': {
                                '$ref': f'#/components/schemas/{func.resp}'
                            }
                        }
                    },
                }
            elif not has_2xx:
                operation['responses']['200'] = {'description': 'Successful Response'}

            if any([hasattr(func, schema) for schema in ('query', 'body')]):
                operation['responses']['400'] = {
                    'description': 'Validation Error',
                    'content': {
                        'application/json': {
                            'schema': {
                                "code": 200,
                            }
                        }
                    },
                }
            routes.setdefault(path, {})[method.lower()] = operation

    definitions = {}
    for _, schema in models.items():
        if 'definitions' in schema:
            for key, value in schema['definitions'].items():
                definitions[key] = value
            del schema['definitions']
    info = {
        'title': title,
        'version': version,
    }
    if description:
        info["description"] = description
    data = {
        'openapi': openapi_version,
        'info': info,
        'tags': list(tags.values()),
        'x-tagGroups': [{"name": k, "tags": list(set(v))} for k, v in groups.items()],
        'paths': {
            **routes
        },
        'components': {
            'schemas': {
                name: schema for name, schema in models.items()
            },
        },
        'definitions': definitions
    }
    return data



import os
from functools import wraps
from typing import Any, Dict, Optional, Type

import pydantic
from flask import Blueprint, Flask, jsonify, render_template, request
from pydantic import BaseModel
from pydantic.typing import Literal

from . import error, openapi, utils
from .error import ValidationError

__all__ = ["CreateDoc", "Docs", "ValidationError"]

__version__ = "1.0.2"

SUPPORTED_UI = ("redoc", "swagger", "rapidoc")


class CreateDoc:
    def __init__(
        self,
        app: Flask = None,
        title: str = "API Docs",
        description: str = "",
        version="latest",
        doc_url: Optional[str] = "/docs",
        openapi_url: Optional[str] = "/openapi.json",
        ui: Literal["redoc", "swagger", "rapidoc"] = "swagger",
        models: Dict[str, Dict] = {}
    ):
        self.app = app
        self._openapi = None
        self.title = title
        self.description = description
        self.version = version
        self.doc_url = doc_url
        self.openapi_url = openapi_url
        self.openapi_version = "3.0.2"
        self.ui = ui
        self.models = models
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask):
        self.app = app
        self._register_doc_blueprint()

    def _register_doc_blueprint(self):
        """
        注册文档蓝图
        """
        template_folder = os.path.join(os.path.dirname(__file__), "templates")
        siwa_bp = Blueprint(
            "siwadoc",
            __name__,
            template_folder=template_folder,
        )

        @siwa_bp.route(self.doc_url)
        def doc_html():
            ui = request.args.get("ui") or self.ui
            assert ui in SUPPORTED_UI, f"ui only support with {SUPPORTED_UI}"
            ui_file = f"{ui}.html"
            return render_template(ui_file, spec_url=self.openapi_url)

        @siwa_bp.route(f"{self.openapi_url}")
        def doc_json():
            return jsonify(self.openapi)

        self.app.register_blueprint(siwa_bp)

    @property
    def openapi(self):
        if not self._openapi:
            self._openapi = openapi.generate_openapi(
                openapi_version=self.openapi_version,
                title=self.title,
                version=self.version,
                description=self.description,
                app=self.app,
                models=self.models,
            )
        return self._openapi

    def doc(
        self,
        query: Optional[Type[BaseModel]] = None,
        header: Optional[Type[BaseModel]] = None,
        cookie: Optional[Type[BaseModel]] = None,
        body: Optional[Type[BaseModel]] = None,
        resp=None,
        x=[],
        tags=[],
        group=None,
        summary=None,
        description=None,
    ):
        """
        装饰器同时兼具文档生成和请求数据校验功能
        """

        def decorate_validate(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                query_data, body_data = None, None
                # 注解参数
                query_in_kwargs = func.__annotations__.get("query")
                body_in_kwargs = func.__annotations__.get("body")
                query_model = query_in_kwargs or query
                body_model = body_in_kwargs or body

                if query_model:
                    query_params = utils.convert_query_params(request.args, query_model)
                    try:
                        query_data = query_model(**query_params)
                    except pydantic.error_wrappers.ValidationError as e:
                        raise ValidationError(e)

                if body_model is not None:
                    try:
                        body_data = body_model(
                            **(request.get_json(force=True, silent=True) or {})
                        )
                    except pydantic.error_wrappers.ValidationError as e:
                        raise ValidationError(e)
                if query_in_kwargs:
                    kwargs["query"] = query_data
                if body_in_kwargs:
                    kwargs["body"] = body_data

                return func(*args, **kwargs)

            for model, name in zip(
                (query, header, cookie, body, resp),
                ("query", "header", "cookie", "body", "resp"),
            ):
                if model:
                    assert issubclass(model, BaseModel)
                    self.models[model.__name__] = model.schema()
                    setattr(wrapper, name, model.__name__)

            code_msg = {}
            if code_msg:
                wrapper.x = code_msg

            if tags:
                wrapper.tags = tags
            if group:
                wrapper.group = group
            wrapper.summary = summary
            wrapper.description = description
            wrapper._decorated = True  # 标记判断改函数是否加入openapi
            return wrapper

        return decorate_validate


class Docs:
    def __init__(self, tags: list[str] = [], group: Optional[str] = None) -> None:
        self.tags = tags
        self.group = group
        self.models: Dict[str, Dict] = {}

    def __call__(
        self,
        query: Optional[Type[BaseModel]] = None,
        header: Optional[Type[BaseModel]] = None,
        cookie: Optional[Type[BaseModel]] = None,
        body: Optional[Type[BaseModel]] = None,
        resp: Optional[Any] = None,
        x=[],
        tags=[],
        group=None,
        summary=None,
        description=None,
    ):
        """
        装饰器同时兼具文档生成和请求数据校验功能
        """

        def decorate_validate(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                query_data, body_data = None, None
                # 注解参数
                query_in_kwargs = func.__annotations__.get("query")
                body_in_kwargs = func.__annotations__.get("body")
                query_model = query_in_kwargs or query
                body_model = body_in_kwargs or body

                if query_model:
                    query_params = utils.convert_query_params(request.args, query_model)
                    try:
                        query_data = query_model(**query_params)
                    except pydantic.error_wrappers.ValidationError as e:
                        raise ValidationError(e)

                if body_model is not None:
                    try:
                        body_data = body_model(
                            **(request.get_json(force=True, silent=True) or {})
                        )
                    except pydantic.error_wrappers.ValidationError as e:
                        raise ValidationError(e)
                if query_in_kwargs:
                    kwargs["query"] = query_data
                if body_in_kwargs:
                    kwargs["body"] = body_data

                return func(*args, **kwargs)

            for model, name in zip(
                (query, header, cookie, body, resp),
                ("query", "header", "cookie", "body", "resp"),
            ):
                if model:
                    assert issubclass(model, BaseModel)
                    self.models[model.__name__] = model.schema()
                    setattr(wrapper, name, model.__name__)

            code_msg = {}
            if code_msg:
                wrapper.x = code_msg

            wrapper.tags = tags or self.tags
            wrapper.group = group or self.group
            wrapper.summary = summary
            wrapper.description = description
            wrapper._decorated = True  # 标记判断改函数是否加入openapi
            return wrapper

        return decorate_validate



from typing import Optional

from pydantic import BaseModel, AnyUrl, EmailStr


class Contact(BaseModel):
    name: Optional[str] = None
    url: Optional[AnyUrl] = None
    email: Optional[EmailStr] = None

    class Config:
        extra = "allow"


class License(BaseModel):
    name: str
    url: Optional[AnyUrl] = None

    class Config:
        extra = "allow"


class Info(BaseModel):
    title: str
    description: Optional[str] = None
    termsOfService: Optional[str] = None
    contact: Optional[Contact] = None
    license: Optional[License] = None
    version: str

    class Config:
        extra = "allow"
